import importlib


class NeomodelException(Exception):
    """
    A base class that identifies all exceptions raised by :mod:`neomodel`.
    """


class AttemptedCardinalityViolation(NeomodelException):
    """
    Attempted to alter the database state against the cardinality definitions.

    Example: a relationship of type `One` trying to connect a second node.
    """
    pass


class CardinalityViolation(NeomodelException):
    """
    The state of database doesn't match the nodes cardinality definition.

    For example a relationship type `OneOrMore` returns no nodes.
    """
    def __init__(self, rel_manager, actual):
        self.rel_manager = str(rel_manager)
        self.actual = str(actual)

    def __str__(self):
        return "CardinalityViolation: Expected: {0}, got: {1}."\
            .format(self.rel_manager, self.actual)


class ModelDefinitionMismatch(NeomodelException):
    """
    Raised when it is impossible to resolve a Neo4j driver Node to a 
    specific object. This can happen as a result of a query returning 
    nodes for which class definitions do exist but have not been loaded 
    (or rather imported) or because the retrieved nodes contain more 
    labels for a known class.
    
    In either of these cases the mismatch must be reported
    """
    
    def __init__(self, db_node, current_node_class_registry):
        """
        Initialises the exception with the database node that failed to resolve 
        to an object
        
        :param db_node: Neo4j driver node object
        :param current_node_class_registry: Dictionary that maps frozenset of
               node labels to model classes
        """
        self.db_node = db_node
        self.current_node_class_registry = current_node_class_registry
        
    def __str__(self):
        node_labels = ",".join(self.db_node.labels)
        ncr_items = list(map(lambda x:"{} --> {}".format(",".join(x[0]), x[1]),
                             self.current_node_class_registry.items()))
        ncr_items_multiline = "\n".join(ncr_items)

        return "Node with labels {} does not resolve to any of the known " \
               "objects\n{}\n".format(node_labels, ncr_items_multiline)


class ConstraintValidationFailed(ValueError, NeomodelException):
    def __init__(self, msg):
        self.message = msg


class DeflateError(ValueError, NeomodelException):
    def __init__(self, key, cls, msg, obj):
        self.property_name = key
        self.node_class = cls
        self.msg = msg
        self.obj = repr(obj)

    def __str__(self):
        return ("Attempting to deflate property '{0}' on {1} of class '{2}': "
                "{3}".format(self.property_name, self.obj,
                             self.node_class.__name__, self.msg))


class DoesNotExist(NeomodelException):
    _model_class = None
    """
    This class property refers the model class that a subclass of this class
    belongs to. It is set by :class:`~neomodel.core.NodeMeta`.
    """

    def __init__(self, msg):
        if self._model_class is None:
            raise RuntimeError("This class hasn't been setup properly.")
        self.message = msg
        super(DoesNotExist, self).__init__(self, msg)

    def __reduce__(self):
        return _unpickle_does_not_exist, (self._model_class, self.message)


def _unpickle_does_not_exist(_model_class, message):
    return _model_class.DoesNotExist(message)


class InflateConflict(NeomodelException):
    def __init__(self, cls, key, value, nid):
        self.cls_name = cls.__name__
        self.property_name = key
        self.value = value
        self.nid = nid

    def __str__(self):
        return """Found conflict with node {0}, has property '{1}' with value '{2}'
            although class {3} already has a property '{1}'""".format(
            self.nid, self.property_name, self.value, self.cls_name)


class InflateError(ValueError, NeomodelException):
    def __init__(self, key, cls, msg, obj=None):
        self.property_name = key
        self.node_class = cls
        self.msg = msg
        self.obj = repr(obj)

    def __str__(self):
        return ("Attempting to inflate property '{0}' on {1} of class '{2}': "
                "{3}".format(self.property_name, self.obj,
                             self.node_class.__name__, self.msg))


class DeflateConflict(InflateConflict):
    def __init__(self, cls, key, value, nid):
        self.cls_name = cls.__name__
        self.property_name = key
        self.value = value
        self.nid = nid if nid else '(unsaved)'

    def __str__(self):
        return """Found trying to set property '{1}' with value '{2}' on node {0}
            although class {3} already has a property '{1}'""".format(
            self.nid, self.property_name, self.value, self.cls_name)


class MultipleNodesReturned(ValueError, NeomodelException):
    def __init__(self, msg):
        self.message = msg


class NotConnected(NeomodelException):
    def __init__(self, action, node1, node2):
        self.action = action
        self.node1 = node1
        self.node2 = node2

    def __str__(self):
        return ("Error performing '{0}' - Node {1} of type '{2}' is not "
                "connected to {2} of type '{3}'."
                .format(self.action, self.node1.id,
                        self.node1.__class__.__name__, self.node2.id,
                        self.node2.__class__.__name__))


class RequiredProperty(NeomodelException):
    def __init__(self, key, cls):
        self.property_name = key
        self.node_class = cls

    def __str__(self):
        return "property '{0}' on objects of class {1}".format(
            self.property_name, self.node_class.__name__)


class UniqueProperty(ConstraintValidationFailed):
    def __init__(self, msg):
        self.message = msg


__all__ = (
    AttemptedCardinalityViolation.__name__,
    CardinalityViolation.__name__,
    ConstraintValidationFailed.__name__,
    DeflateConflict.__name__,
    DeflateError.__name__,
    DoesNotExist.__name__,
    InflateConflict.__name__,
    InflateError.__name__,
    MultipleNodesReturned.__name__,
    NeomodelException.__name__,
    NotConnected.__name__,
    RequiredProperty.__name__,
    UniqueProperty.__name__
)



import os
import logging

from neomodel import db, config
from neomodel.match_q import Q
from neomodel.match import NodeSet
from neomodel.cardinality import ZeroOrOne
from neomodel.relationship_manager import ZeroOrMore
from neomodel.relationship_manager import RelationshipFrom
from neomodel.relationship_manager import RelationshipTo
from neomodel.relationship import StructuredRel
from neomodel import (
    StructuredNode,
    StringProperty,
    DateProperty,
    DateTimeProperty,
    install_all_labels,
)

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)


class ItemCarRel(StructuredRel):
    name = StringProperty()


class User(StructuredNode):
    username = StringProperty()
    car = RelationshipFrom('Car', 'OWN', ZeroOrMore)


class Item(StructuredNode):
    car = RelationshipFrom('Car', 'U', ZeroOrMore, model=ItemCarRel)

class Car(StructuredNode):
    name = StringProperty()
    owner = RelationshipTo('User', 'OWN', ZeroOrOne)
    item = RelationshipTo('Item', 'U', ZeroOrMore, model=ItemCarRel)

    datetime_auto_update = DateTimeProperty(default_now=True, auto_update=True)


def delete(model):
    db.cypher_query("""
    MATCH (m:{model})
    DETACH DELETE m;
    """.format(model=model.__name__))


def test_has_functionality():
    delete(Car)
    delete(User)
    delete(Item)
    m = Car(name='maserrati').save()
    l = Car(name='lamborgini').save()
    h = Car(name='hyndau').save()

    item = Item().save()
    m.item.connect(item)
    m.item.connect(Item().save())

    u1 = User(username="Jack").save()
    u1.car.connect(m)

    u2 = User(username="Sasha").save()
    u2.car.connect(h)
    u2.car.connect(l)


    users = User.nodes.filter(username__in=["Jack", 'Sasha'])
    cars = Car.nodes.has(owner=users)
    items = Item.nodes.filter()
    cars = cars.has(item=items)
    print(cars.all())

    user = User.nodes.first()

    user_cars = user.car.filter()
    user_cars = user_cars.has(item=Item.nodes.filter())
    print(list(user_cars))

    users = User.nodes.filter()
    cars = Car.nodes.has(item=items, is_not=True)
    cars = cars.has(owner=users, is_not=True).distinct()
    print(list(cars))


    user.car.connect_nodeset(Car.nodes.filter())

    #  print(list(users))

    #  print(len(Car.nodes.filter(Q(item__name__icontains='fire') | Q(item__name__contains='asdsad') | Q(name__icontains='maser')).distinct().return_fields(['car_item', ])))
    raise Exception("stop")


    #  print(Car.nodes.filter(Q()).all())

    #  print(Car.nodes.has(owner=u2).all())


    query = """
    match (c:Car)
        WHERE c.uid = $car_uid
    WITH c
    """
    params = {'car_uid': '2'}
    cars = Car.nodes.extend_cypher(query, params).filter(name__icontains='asdasd').all()


    query = """
    CALL apoc.cypher.run("
        MATCH (i:Item)
            WHERE i.name = $swag
        RETURN i
    ", $params) YIELD value
    WITH value.i as item
    RETURN item;
    """
    results, _ = db.cypher_query(query, {'params': {'swag': "asdasd"}})
    print(results)

    cars_union = Car.nodes.union(item=Item.nodes.filter(car__name__exact='asdasd'), \
            owner=User.nodes.has(car=Car.nodes.filter())).filter(name__icontains='asdasd')
    print(cars_union.all())

    #  car = Car.nodes.filter()
    #  car = car.return_fields(['car', ])
    #  print(len(car))
    #  print(len(car[0].owner.all_relationships()))
    #  print(len(car[1].item.filter_relationships(Item.nodes.first(),
    #      Q(Q(name__icontains='asdasd') | Q(name__icontains='asdasxc')))))

    #  print(ItemCarRel.defined_properties(rels=False))



if __name__ == "__main__":
    config.DATABASE_URL = os.environ.get('NEO4J_BOLT_URL',
                                         'bolt://neo4j:neo4j_admin@localhost:7687')
    config.AUTO_INSTALL_LABELS = True
    install_all_labels()

    # RUN LOCAL TESTS
    test_has_functionality()
#      db.cypher_query("""
#  MATCH (user:User) WITH user
#  MATCH (user)--(a:Car) return user as u_union
#  UNION
#  MATCH (user:User) WITH user
#  MATCH (user)--(b:Car) return user as u_union""")



# encoding=utf-8
import inspect
import re

from .core import StructuredNode, db
from .properties import AliasProperty
from .exceptions import MultipleNodesReturned, NeomodelException
from .match_q import Q, QBase
from functools import singledispatch, reduce
from .util import get_rhs_ident, UnionBlock

OUTGOING, INCOMING, EITHER = 1, -1, 0


# basestring python 3.x fallback
try:
    basestring
except NameError:
    basestring = str


def _rel_helper(
        lhs,
        rhs,
        ident=None,
        relation_type=None,
        direction=None,
        relation_properties=None,
        **kwargs):
    """
    Generate a relationship matching string, with specified parameters.
    Examples:
    relation_direction = OUTGOING: (lhs)-[relation_ident:relation_type]->(rhs)
    relation_direction = INCOMING: (lhs)<-[relation_ident:relation_type]-(rhs)
    relation_direction = EITHER: (lhs)-[relation_ident:relation_type]-(rhs)

    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param relation_properties: dictionary of relationship properties to match
    :returns: string
    """
    if kwargs.get('reverse'):
        direction = -direction

    if direction == OUTGOING:
        stmt = '-{0}->'
    elif direction == INCOMING:
        stmt = '<-{0}-'
    else:
        stmt = '-{0}-'

    rel_props = ''

    if relation_properties:
        rel_props = ' {{{0}}}'.format(', '.join(
            ['{}: {}'.format(key, value) for key, value in relation_properties.items()]))

    # direct, relation_type=None is unspecified, relation_type
    if relation_type is None:
        stmt = stmt.format('')
    # all("*" wildcard) relation_type
    elif relation_type == '*':
        stmt = stmt.format('[*]')
    else:
        # explicit relation_type
        stmt = stmt.format('[%s:`%s`%s]' % (ident if ident else '', relation_type, rel_props))

    return "({0}){1}({2})".format(lhs, stmt, rhs)


# special operators
_SPECIAL_OPERATOR_IN = 'IN'
_SPECIAL_OPERATOR_INSENSITIVE = '(?ui)'
_SPECIAL_OPERATOR_ISNULL = 'IS NULL'
_SPECIAL_OPERATOR_ISNOTNULL = 'IS NOT NULL'
_SPECIAL_OPERATOR_REGEX = '=~'

_UNARY_OPERATORS = (_SPECIAL_OPERATOR_ISNULL, _SPECIAL_OPERATOR_ISNOTNULL)

_REGEX_INSESITIVE = _SPECIAL_OPERATOR_INSENSITIVE + '{}'
_REGEX_CONTAINS = '.*{}.*'
_REGEX_STARTSWITH = '{}.*'
_REGEX_ENDSWITH = '.*{}'

# regex operations that require escaping
_STRING_REGEX_OPERATOR_TABLE = {
    'iexact': _REGEX_INSESITIVE,
    'contains': _REGEX_CONTAINS,
    'icontains': _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_CONTAINS,
    'startswith': _REGEX_STARTSWITH,
    'istartswith': _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_STARTSWITH,
    'endswith': _REGEX_ENDSWITH,
    'iendswith': _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_ENDSWITH,
}
# regex operations that do not require escaping
_REGEX_OPERATOR_TABLE = {
    'iregex': _REGEX_INSESITIVE,
}
# list all regex operations, these will require formatting of the value
_REGEX_OPERATOR_TABLE.update(_STRING_REGEX_OPERATOR_TABLE)

# list all supported operators
OPERATOR_TABLE = {
    'lt': '<',
    'gt': '>',
    'lte': '<=',
    'gte': '>=',
    'ne': '<>',
    'in': _SPECIAL_OPERATOR_IN,
    'isnull': _SPECIAL_OPERATOR_ISNULL,
    'regex': _SPECIAL_OPERATOR_REGEX,
    'exact': '='
}
# add all regex operators
OPERATOR_TABLE.update(_REGEX_OPERATOR_TABLE)


def install_traversals(cls, node_set):
    """
    For a StructuredNode class install Traversal objects for each
    relationship definition on a NodeSet instance
    """
    rels = cls.defined_properties(rels=True, aliases=False, properties=False)

    for key, value in rels.items():
        if hasattr(node_set, key):
            raise ValueError("Can't install traversal '{}' exists on NodeSet".format(key))

        rel = getattr(cls, key)
        rel._lookup_node_class()

        traversal = Traversal(source=node_set, name=key, definition=rel.definition)
        setattr(node_set, key, traversal)


re._alphanum_str = frozenset(
    "_абвгдеёжзийклмнопрстуфхцчшщъыьэюяАБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ\
            abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890")


def preprocess_filter_find_args(cls, kwargs):
    # realisation
    key, value = list(kwargs.items())[0]

    if len(re.findall(r"__", key)) != 2:
        raise Exception("Not found the double \'__\'")

    cuttener = key.find("__")
    relationship_field = key[0:cuttener]
    key = key[cuttener + 2:]
    del cuttener
    if not hasattr(cls, relationship_field):
        raise Exception("field does not exists in source fucking idiot")

    rel_field = getattr(cls, relationship_field)
    kwargs = {key: value}
    res = process_filter_args(rel_field.definition['model'], kwargs)
    # returns where definition
    return res, rel_field


def process_filter_args(cls, kwargs):
    """
    loop through properties in filter parameters check they match class definition
    deflate them and convert into something easy to generate cypher from
    """

    output = {}

    for key, value in kwargs.items():
        if '__' in key:
            prop, operator = key.rsplit('__')
            operator = OPERATOR_TABLE[operator]
        else:
            prop = key
            operator = '='

        if prop not in cls.defined_properties(rels=False):
            raise ValueError("No such property {} on {}".format(prop, cls.__name__))

        property_obj = getattr(cls, prop)
        if isinstance(property_obj, AliasProperty):
            prop = property_obj.aliased_to()
            deflated_value = getattr(cls, prop).deflate(value)
        else:
            # handle special operators
            if operator == _SPECIAL_OPERATOR_IN:
                if not isinstance(value, tuple) and not isinstance(value, list):
                    raise ValueError(
                        'Value must be a tuple or list for IN operation {}={}'.format(key, value))
                deflated_value = [property_obj.deflate(v) for v in value]
            elif operator == _SPECIAL_OPERATOR_ISNULL:
                if not isinstance(value, bool):
                    raise ValueError('Value must be a bool for isnull operation on {}'.format(key))
                operator = 'IS NULL' if value else 'IS NOT NULL'
                deflated_value = None
            elif operator in _REGEX_OPERATOR_TABLE.values():
                deflated_value = property_obj.deflate(value)
                if not isinstance(deflated_value, basestring):
                    raise ValueError('Must be a string value for {}'.format(key))
                if operator in _STRING_REGEX_OPERATOR_TABLE.values():
                    deflated_value = re.escape(deflated_value)
                deflated_value = operator.format(deflated_value)
                operator = _SPECIAL_OPERATOR_REGEX
            else:
                deflated_value = property_obj.deflate(value)

        # map property to correct property name in the database
        db_property = cls.defined_properties(rels=False)[prop].db_property or prop

        output[db_property] = (operator, deflated_value)
    return output


def process_has_args(builder, kwargs, is_not):
    """
    loop through has parameters check they correspond to class rels defined
    """
    # nodeset
    cls = builder.source_class
    # all relationships from model
    rel_definitions = cls.defined_properties(properties=False, rels=True, aliases=False)

    for key, value in kwargs.items():
        if key not in rel_definitions:
            raise ValueError("No such relation {} defined on a {}".format(key, cls.__name__))

        # initialize lookup node class of `key` field from source_class
        rel_definitions[key]._lookup_node_class()
        # get `key` definition from field
        definition = rel_definitions[key].definition
        # set parameters for build class, for create relationship query
        update_matches(value, builder, key, definition, is_not, **kwargs)


def process_union(nodeset, kwargs):
    """
    Attrs:
        nodeset: source nodeset
        kwargs - params
    """
    assert len(kwargs.items()) > 1, 'Union function may union greater or equal than 2 nodesets'

    # source model
    cls = nodeset.source_class
    # all relationships from model
    rel_definitions = cls.defined_properties(properties=False, rels=True, aliases=False)

    for key, value in kwargs.items():
        if key not in rel_definitions:
            raise ValueError("No such relation {} defined on a {}".format(key, cls.__name__))
        # initialize lookup node class of `key` field from source_class
        rel_definitions[key]._lookup_node_class()
        # get `key` definition from field
        definition = rel_definitions[key].definition
        union_block = UnionBlock(nodeset, key, definition, kwargs[key], source_model=cls)
        nodeset.union_operations.append(union_block)


@singledispatch
def update_matches(argument, builder, key, definition, *args, **kwargs):
    raise NotImplementedError("Type {} not implemented yet".format(type(argument)))


@singledispatch
def generate_label(type, **kwargs):
    raise NotImplementedError(
        "Label of type: {}, not implemented yet.".format(
            type(
                kwargs['value'])))


@generate_label.register(StructuredNode)
def _generate_label_sn(type, **kwargs):
    self = kwargs['self']
    supplicant = kwargs['label'].lower()[1:] + "_uid"
    label = kwargs['label'] + ' {{ uid: ${}}}'.format(supplicant)
    where_relation = _rel_helper(
        lhs=kwargs['source_ident'],
        rhs=label,
        ident='',
        **kwargs['val'])
    self._ast['where'].append(where_relation)
    self._query_params.update({supplicant: kwargs['value'].uid})
    return self


@update_matches.register(bool)
def _um_register(argument, builder, key, definition, *args, **kwargs):
    """
    default match
    """
    if argument:
        builder.must_match[key] = {'definition': definition}
    else:
        builder.dont_match[key] = {'definition': definition}


class QueryBuilder(object):

    def __init__(self, node_set):
        self.node_set = node_set
        self._ast = {
            'match': [],
            'where': [],
            'lookup': [],
            'distinct': node_set.is_distinct,
        }
        if hasattr(node_set, 'lookup') and node_set.lookup is not None:
            self._ast['lookup'].append(node_set.lookup)
        self._query_params = {}
        if hasattr(node_set, '_query_params'):
            self._query_params = node_set._query_params
        self._place_holder_registry = {}
        self._ident_count = 0

        # no label in ident
        self.no_label_in_ident = False
        if hasattr(node_set, 'no_label_in_ident'):
            self.no_label_in_ident = node_set.no_label_in_ident

        # custom returnable fields
        self._return_fields = []
        if hasattr(node_set, '_return_fields'):
            self._return_fields = ", ".join(node_set._return_fields)

    def build_ast(self):
        self.build_source(self.node_set)

        if hasattr(self.node_set, '_extra_queries') and self.node_set._extra_queries['need']:
            self.build_extra_ast(self.node_set._extra_queries)

        if getattr(self.node_set, 'skip', None) is not None:
            self._ast['skip'] = self.node_set.skip
        if getattr(self.node_set, 'limit', None) is not None:
            self._ast['limit'] = self.node_set.limit
        return self

    def build_extra_ast(self, extra):
        quote = ','
        if extra['match']:
            self._ast['match'] = list(set(self._ast['match']) | set(extra['match']))
        if extra.get('return') is not None:
            self._ast['return'] = extra['return']

    def build_relationship_filters(self, ident, filters, source_class):
        if filters is not None and isinstance(filters, QBase):
            match_stmt, where_stmt = self._parse_q_find_filters(ident, filters, source_class, [])
            match_stmt = set(match_stmt)
            if match_stmt:
                self._ast['match'] += list(match_stmt)
            self._ast['where'].append(where_stmt)

    def _parse_q_find_filters(self, ident, q, source_class, matches=[]):
        target = []
        for child in q.children:
            if isinstance(child, QBase):
                matches, q_childs = self._parse_q_find_filters(ident, child, source_class, matches)
                if child.connector == Q.OR:
                    q_childs = "(" + q_childs + ")"
                target.append(q_childs)
            else:
                kwargs = {child[0]: child[1]}

                # match relationship filters
                filters, rel_field = preprocess_filter_find_args(source_class, kwargs)

                rhs_ident = get_rhs_ident(rel_field)
                rel_ident = rhs_ident.lower()

                matches.append(
                    _rel_helper(
                        ident,
                        ':' + rhs_ident,
                        rel_ident,
                        rel_field.definition['relation_type']))

                # relationship where statement
                for prop, op_and_val in filters.items():
                    op, val = op_and_val
                    if op in _UNARY_OPERATORS:
                        # unary operators do not have a parameter
                        statement = '{}.{} {}'.format(rel_ident, prop, op)
                    else:
                        place_holder = self._register_place_holder(rel_ident + '_' + prop)
                        statement = '{}.{} {} {{{}}}'.format(rel_ident, prop, op, place_holder)
                        self._query_params[place_holder] = val
                    target.append(statement)
        ret = ' {} '.format(q.connector).join(target)
        if q.negated:
            ret = 'NOT ({})'.format(ret)
        return matches, ret

    def build_source(self, source):
        if isinstance(source, (Traversal, TraversalRelationships)):
            return self.build_traversal(source)
        elif isinstance(source, NodeSet):

            if inspect.isclass(source.source) and issubclass(source.source, StructuredNode):
                attrs = (source.ident, source)
                self.generate_union_query(*attrs)
                ident = self.build_label(*attrs)
            else:
                ident = self.build_source(source.source)

            # has, union calls
            self.build_additional_match(ident, source)

            # order by filtering
            if hasattr(source, '_order_by'):
                self.build_order_by(ident, source)

            # base filtering
            if source.filters or source.q_filters:
                self.build_where_stmt(
                    ident,
                    source.filters,
                    source.q_filters,
                    source_class=source.source_class)

            if source.relationship_filters:
                self.build_relationship_filters(ident,
                                                source.relationship_filters,
                                                source.source_class)

            return ident
        elif isinstance(source, StructuredNode):
            return self.build_node(source)
        else:
            raise ValueError("Unknown source type " + repr(source))

    def create_ident(self):
        self._ident_count += 1
        return 'r' + str(self._ident_count)

    def build_order_by(self, ident, source):
        if '?' in source._order_by:
            self._ast['with'] = '{}, rand() as r'.format(ident)
            self._ast['order_by'] = 'r'
        else:
            self._ast['order_by'] = []
            for p in source._order_by:
                if p['ident']:
                    self._ast['order_by'].append('{}.{}'.format(ident, p['name'] + p['type']))
                else:
                    self._ast['order_by'].append('{}'.format(p['name'] + p['type']))

    def build_traversal(self, traversal):
        """
        traverse a relationship from a node to a set of nodes
        """
        # build source
        lhs_ident = self.build_source(traversal.source)
        rhs_ident = "{alias}:{def_label}".format(
            alias=traversal.name,
            def_label=traversal.definition.get('node_class').__label__)

        if isinstance(traversal, TraversalRelationships):
            self._ast['return'] = traversal.rel_ident
            rel_ident = traversal.rel_ident
        elif isinstance(traversal, Traversal):
            self._ast['return'] = traversal.name
            rel_ident = self.create_ident()

        self._ast['result_class'] = traversal.target_class

        # create static rel ident
        stmt = _rel_helper(lhs=lhs_ident, rhs=rhs_ident, ident=rel_ident, **traversal.definition)
        self._ast['match'].append(stmt)

        if traversal.filters:
            self.build_where_stmt(rel_ident, traversal.filters)
        return traversal.name

    def build_node(self, node):
        ident = node.__class__.__name__.lower()
        place_holder = self._register_place_holder(ident)

        # Hack to emulate START to lookup a node by id
        _node_lookup = 'MATCH ({}) WHERE id({})={{{}}} WITH {}'.format(
            ident, ident, place_holder, ident)
        self._ast['lookup'].append(_node_lookup)

        self._query_params[place_holder] = node.id

        self._ast['return'] = ident
        self._ast['result_class'] = node.__class__
        return ident

    def build_label(self, ident, nodeset):
        """
        match nodes by a label
        """
        ident_w_label = ident if self.no_label_in_ident else \
            ident + ':' + nodeset.source.__label__
        self._ast['match'].append('({})'.format(ident_w_label))
        if self._return_fields:
            self._ast['return'] = ident + ', ' + self._return_fields
        else:
            self._ast['return'] = ident
        self._ast['result_class'] = nodeset.source
        return ident

    def generate_query(self, ident, node_set, attr='must_match'):
        for key, value in getattr(node_set, attr).items():
            if isinstance(value, dict):
                val = value['definition']
                label = ':' + val['node_class'].__label__
                if attr == 'extra_match':
                    # insert value definitions in match
                    # value['value'] = nodeset in has keyword arguments
                    # key - keyword in has argument
                    # value - definition set
                    # label - Relationship key definition field label
                    # val - Relationship definition ('node_class', 'label', ...)
                    # node_set - self nodeset
                    props = {
                        'val': val,
                        'source_ident': ident,
                        'label': label,
                        'value': value['value'],
                        'nodeset_origin': node_set,
                        'self': self,
                        'is_not': value.get('is_not', False),
                    }
                    generate_label(value['value'], **props)
                else:
                    where_rel = _rel_helper(lhs=ident, rhs=label, ident='', **val)
                    stmt = [where_rel, ] if attr == 'must_match' else ['NOT ' + where_rel, ]
                    self._ast['where'] += stmt
            else:
                raise ValueError("Expecting dict got: " + repr(value))

    def generate_union_query(self, ident, nodeset):
        if nodeset.union_operations:
            self.no_label_in_ident = True

            query, params = UnionBlock.union_queries(ident, nodeset.union_operations)

            # extend cypher realisation
            self._ast['lookup'].append(query)
            self._query_params.update(params)

    def build_additional_match(self, ident, node_set):
        """handle additional matches supplied by 'has()' calls
        """
        self.generate_query(ident, node_set)
        self.generate_query(ident, node_set, 'dont_match')
        self.generate_query(ident, node_set, 'extra_match')

    def _register_place_holder(self, key):
        self._place_holder_registry[key] = self._place_holder_registry.get(key, 0) + 1
        return key + '_' + str(self._place_holder_registry[key])

    def _where_query(self, ident, q, source_class):
        match_stmt, where_stmt = self._parse_q_filters(ident, q, source_class)
        match_stmt = set(match_stmt)
        if match_stmt:
            self._ast['match'] += list(match_stmt)
        if where_stmt:
            self._ast['where'].append(where_stmt)
        return where_stmt, self._query_params

    def _parse_q_filters(self, ident, q, source_class, matches=[]):
        target = []
        for child in q.children:
            if isinstance(child, QBase):
                matches, q_childs = self._parse_q_filters(ident, child, source_class, matches)
                if child.connector == Q.OR:
                    q_childs = "(" + q_childs + ")"
                target.append(q_childs)
            else:
                # for identation of relationship fields or not
                operator = ident
                kwargs = {child[0]: child[1]}
                if len(child[0].split('__')) == 3:
                    # find_by_edge
                    filters, rel_field = preprocess_filter_find_args(source_class, kwargs)

                    rhs_ident = get_rhs_ident(rel_field)
                    rel_ident = "{}_{}".format(ident, rhs_ident.lower())

                    matches.append(
                        _rel_helper(
                            ident,
                            ':' + rhs_ident,
                            rel_ident,
                            rel_field.definition['relation_type']))
                    operator = rel_ident
                else:
                    # base filter
                    filters = process_filter_args(source_class, kwargs)
                for prop, op_and_val in filters.items():
                    op, val = op_and_val
                    if op in _UNARY_OPERATORS:
                        # unary operators do not have a parameter
                        statement = '{}.{} {}'.format(operator, prop, op)
                    else:
                        place_holder = self._register_place_holder(operator + '_' + prop)
                        statement = '{}.{} {} {{{}}}'.format(operator, prop, op, place_holder)
                        self._query_params[place_holder] = val
                    target.append(statement)
        ret = ' {} '.format(q.connector).join(target)
        if q.negated:
            ret = 'NOT ({})'.format(ret)
        return matches, ret

    def build_where_stmt(self, ident, filters, q_filters=None, source_class=None):
        """
        construct a where statement from some filters
        """
        if q_filters is not None:
            match_stmt, where_stmt = self._parse_q_filters(ident, q_filters, source_class, [])
            match_stmt = set(match_stmt)
            if match_stmt:
                self._ast['match'] += list(match_stmt)
            if where_stmt:
                self._ast['where'].append(where_stmt)
        else:
            # DEPRECATED
            stmts = []
            for row in filters:
                negate = False

                # pre-process NOT cases as they are nested dicts
                if '__NOT__' in row and len(row) == 1:
                    negate = True
                    row = row['__NOT__']

                for prop, op_and_val in row.items():
                    op, val = op_and_val
                    if op in _UNARY_OPERATORS:
                        # unary operators do not have a parameter
                        statement = '{} {}.{} {}'.format('NOT' if negate else '', ident, prop, op)
                    else:
                        place_holder = self._register_place_holder(ident + '_' + prop)
                        statement = '{} {}.{} {} {{{}}}'.format(
                            'NOT' if negate else '', ident, prop, op, place_holder)
                        self._query_params[place_holder] = val
                    stmts.append(statement)

            self._ast['where'].append(' AND '.join(stmts))

    def build_query(self, return_operation='RETURN'):
        query = self.build_query_build_lookup()
        query = self.build_query_build_match(query)
        query = self.build_query_build_where(query)
        query = self.build_query_build_with(query)
        query = self.build_query_build_return(query, return_operation)
        return query

    @classmethod
    def gather_lookup_identations(cls, lookups):
        identations = ""
        for indice, lookup_query in enumerate(lookups):
            if indice > 0:
                # append after first lookup query
                lookup_query += identations
            identations += ', ' + lookup_query.strip(' ').split(" ")[-1]
            yield lookup_query

    def build_query_build_lookup(self, query=''):
        if 'lookup' in self._ast and self._ast['lookup']:
            lookups = filter(lambda lookup: lookup, self._ast['lookup'])
            lookups = self.gather_lookup_identations(lookups)
            query += "\n".join(lookups)
        return query

    def build_query_build_match(self, query=''):
        query += ' MATCH '
        query += ', '.join(['({})'.format(i) for i in self._ast['match']])
        return query

    def build_query_build_where(self, query='', **params):
        if 'where' in self._ast and self._ast['where']:
            if params.get('where', True):
                query += ' WHERE '
            query += ' AND '.join(self._ast['where'])
        return query

    def build_query_build_with(self, query=''):
        if 'with' in self._ast and self._ast['with']:
            query += ' WITH '
            query += self._ast['with']
        return query

    def build_query_build_return(self, query='', return_operation=None):
        # set WITH or RETURN value
        query += ' {} '.format(return_operation)
        # set DISTINCT or ''
        if self._ast['distinct']:
            query += "DISTINCT "
        # return ident
        query += self._ast['return']
        if 'order_by' in self._ast and self._ast['order_by']:
            query += ' ORDER BY '
            query += ', '.join(self._ast['order_by'])
        if 'skip' in self._ast:
            query += ' SKIP {0:d}'.format(self._ast['skip'])
        if 'limit' in self._ast:
            query += ' LIMIT {0:d}'.format(self._ast['limit'])
        return query

    def _count(self):
        return_operation = "RETURN"
        return_string = self._ast['return']
        if ',' in return_string:
            return_operation = 'WITH'
            self._ast['return'] = return_string + " RETURN count(*)"
        else:
            self._ast['return'] = 'count({})'.format(return_string)

        #  self._ast['return'] = 'count({})'.format("*")

        # drop order_by, results in an invalid query
        self._ast.pop('order_by', None)
        # drop limit & offset
        limit = self._ast.pop('limit', None)
        skip = self._ast.pop('skip', None)
        query = self.build_query(return_operation)
        results, _ = db.cypher_query(query, self._query_params)
        # update it
        self._ast['limit'] = limit
        self._ast['skip'] = skip
        return int(results[0][0])

    def _contains(self, node_id):
        # inject id = into ast
        ident = self._ast['return']
        place_holder = self._register_place_holder(ident + '_contains')
        self._ast['where'].append('id({}) = {{{}}}'.format(ident, place_holder))
        self._query_params[place_holder] = node_id
        return self._count() >= 1

    def _execute(self):
        query = self.build_query()
        results, _ = db.cypher_query(query, self._query_params, resolve_objects=True)
        # The following is not as elegant as it could be but had to be copied from the
        # version prior to cypher_query with the resolve_objects capability.
        # It seems that certain calls are only supposed to be focusing to the first
        # result item returned (?)
        if results:
            return [n[0] for n in results]
        return []

    def _execute_multiple(self, resolve_objects=True):
        query = self.build_query()
        results, _ = db.cypher_query(query,
                                     self._query_params,
                                     resolve_objects=resolve_objects)
        if results:
            return results
        return []


class BaseSet(object):
    """
    Base class for all node sets.

    Contains common python magic methods, __len__, __contains__ etc
    """
    query_cls = QueryBuilder

    def __init__(self, *args, **kwargs):
        self.is_distinct = False
        #  order_by = [
        #    {
        #      "field_name": "",
        #      "type": "-",
        #      "identation": True
        #    },
        #  ]
        self._order_by = []

    def distinct(self, f=True):
        """
        Return DISTINCT results
        """
        self.is_distinct = f
        return self

    def all(self, multiple=False):
        """
        Return all nodes belonging to the set
        :return: list of nodes
        :rtype: list
        """
        qb = self.query_cls(self).build_ast()
        if multiple:
            return qb._execute_multiple()
        return qb._execute()

    @property
    def ident(self):
        return self.source.__label__.lower()

    def __iter__(self):
        return (i for i in self.query_cls(self).build_ast()._execute())

    def __len__(self):
        return self.query_cls(self).build_ast()._count()

    def __bool__(self):
        return self.query_cls(self).build_ast()._count() > 0

    def __nonzero__(self):
        return self.query_cls(self).build_ast()._count() > 0

    def __contains__(self, obj):
        if isinstance(obj, StructuredNode):
            if hasattr(obj, 'id'):
                return self.query_cls(self).build_ast()._contains(int(obj.id))
            raise ValueError("Unsaved node: " + repr(obj))
        else:
            raise ValueError("Expecting StructuredNode instance")

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.stop and key.start:
                self.limit = key.stop - key.start
                self.skip = key.start
            elif key.stop:
                self.limit = key.stop
            elif key.start:
                self.skip = key.start

            return self.query_cls(self).build_ast()._execute()

        elif isinstance(key, int):
            self.skip = key
            self.limit = 1

            return self.query_cls(self).build_ast()._execute()[0]


class NodeSet(BaseSet):
    """
    A class representing as set of nodes matching common query parameters
    """

    def __init__(self, source, *args, **kwargs):
        super(NodeSet, self).__init__(*args, **kwargs)
        self.source = source  # could be a Traverse object or a node class
        if isinstance(source, Traversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, StructuredNode):
            self.source_class = source
        elif isinstance(source, StructuredNode):
            self.source_class = source.__class__
        else:
            raise ValueError("Bad source for nodeset " + repr(source))

        # setup Traversal objects using relationship definitions
        install_traversals(self.source_class, self)
        self.implicit_init()

    def implicit_init(self,):
        # lazy query builds
        # Fields: return, match, where
        self._extra_queries = {
            'match': [],
            'where': [],
            'need': False,
        }
        # filters
        self.filters = []
        self.q_filters = Q()
        self.relationship_filters = Q()

        # used by has()
        self.must_match = {}
        self.dont_match = {}
        # for nodeset or structured node relationships
        self.extra_match = {}

        # for extend cypher query (from query, to NodeSet)
        self.no_label_in_ident = False
        self._query_params = {}
        self.lookup = ""

        # for union operations (List[Type[UnionBlock]])
        self.union_operations = []

        # for return aliases
        self._return_fields = list()

    def extend_cypher(self, query, params={}):
        """
        Attrs:
            query (str): Cypher query.
            params (dict): Cypher params in `query` param
            query_ident: ident that placed in `query` after WITH predicate
        Returns:
            <NodeSet>
        """
        assert isinstance(query, str), 'query should be str instance'
        assert isinstance(params, dict), 'params should be dict instance'

        self.no_label_in_ident = True

        assert 'WITH' in query, 'query should contains and returns WITH statement'

        query += " AS {}".format(self.ident)

        self.lookup += query
        self._query_params.update(params)
        return self

    def return_fields(self, fields):
        assert isinstance(fields, list), "Fields should be the list type"
        self._return_fields += fields
        return self

    def to_q_filters(self, filters):
        return reduce(lambda initial, n: initial & Q(**{n[0]: n[1]}),
                      filters.items(), Q())

    def set_limit(self, limit=200):
        """
        Lazy limit
        Attrs:
            limit: int (default=200): if None provided, default used
        Returns:
            <NodeSet>
        """
        self.limit = limit
        return self

    def set_skip(self, skip=0):
        """
        Lazy skip
        Attrs:
            skip: int (default=0): if None provided, default used
        Returns:
        """
        self.skip = skip
        return self

    @property
    def query_builder(self, ):
        """
        Return <QueryBuilder> instance.
        """
        return self.query_cls(self).build_ast()

    def _get(self, limit=None, **kwargs):
        self.filter(**kwargs)
        if limit:
            self.limit = limit
        return self.query_cls(self).build_ast()._execute()

    def get(self, **kwargs):
        """
        Retrieve one node from the set matching supplied parameters

        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = self._get(limit=2, **kwargs)
        if len(result) > 1:
            raise MultipleNodesReturned(repr(kwargs))
        elif not result:
            raise self.source_class.DoesNotExist(repr(kwargs))
        else:
            return result[0]

    def get_or_none(self, **kwargs):
        """
        Retrieve a node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return self.get(**kwargs)
        except self.source_class.DoesNotExist:
            pass

    def first(self, **kwargs):
        """
        Retrieve the first node from the set matching supplied parameters

        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = result = self._get(limit=1, **kwargs)
        if result:
            return result[0]
        else:
            raise self.source_class.DoesNotExist(repr(kwargs))

    def first_or_none(self, **kwargs):
        """
        Retrieve the first node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return self.first(**kwargs)
        except self.source_class.DoesNotExist:
            pass

    def filter(self, *args, **kwargs):
        """
        Apply filters to the existing nodes in the set.

        :param kwargs: filter parameters

            Filters mimic Django's syntax with the double '__' to separate field and operators.

            e.g `.filter(salary__gt=20000)` results in `salary > 20000`.

            The following operators are available:

             * 'lt': less than
             * 'gt': greater than
             * 'lte': less than or equal to
             * 'gte': greater than or equal to
             * 'ne': not equal to
             * 'in': matches one of list (or tuple)
             * 'isnull': is null
             * 'regex': matches supplied regex (neo4j regex format)
             * 'exact': exactly match string (just '=')
             * 'iexact': case insensitive match string
             * 'contains': contains string
             * 'icontains': case insensitive contains
             * 'startswith': string starts with
             * 'istartswith': case insensitive string starts with
             * 'endswith': string ends with
             * 'iendswith': case insensitive string ends with

        :return: self
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & Q(*args, **kwargs))
        return self

    def exclude(self, *args, **kwargs):
        """
        Exclude nodes from the NodeSet via filters.

        :param kwargs: filter parameters see syntax for the filter method
        :return: self
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & ~Q(*args, **kwargs))
        return self

    def has(self, **kwargs):
        """
        Nodeset, object, bool types
        Supports operation
         & or |
        """
        is_not = kwargs.pop('is_not', False)
        process_has_args(self, kwargs, is_not=is_not)
        return self

    def union(self, **kwargs):
        process_union(self, kwargs)
        return self

    def find_by_edge(self, *args, **kwargs):
        """
        Extra filter for find nodes by relationship properties
        Example:
            .find_by_edge(Q(research__ru_name__icontains='value'))

            research - relationship definition in our model of nodeset,
                or relationship field at simple
            ru_name - field on edge (defined in relationship_model=in relationship
                    definition)
            icontains - search param
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & Q(*args, **kwargs))
        return self

    def append_relationship(self, field):
        """Field must be relationship definition
        Just append relationship to model. Without extra queries:
            Example:
                coffee.append_relationship('sugar') ->
            MATCH (coffee:Coffee), ((coffee)-[sugar:SugarCoffeeRel]-(sugar_node:Sugar)) ...
        """
        field = getattr(self.source_class, field, None)
        if field is None or not hasattr(field, 'definition'):
            raise NeomodelException("Field {} not found in source_class".format(field))
        if isinstance(self.source, Traversal):
            ident = self.source.target_class.__label__.lower()
        else:
            ident = self.ident
        rhs_ident = get_rhs_ident(field)
        rel_ident = "{}_{}".format(ident, rhs_ident.lower())
        rel_query = _rel_helper(ident,
                                ":{label}".format(label=rhs_ident),
                                rel_ident,
                                field.definition['relation_type'])
        extra_match_set = set(self._extra_queries['match'])
        extra_match_set |= set((rel_query, ))
        self._extra_queries['match'] = list(extra_match_set)
        self._extra_queries['need'] = True
        return self

    def set_return(self, string):
        """
        Warning: request may be incorrect
        """
        self._extra_queries['return'] = string
        self._extra_queries['need'] = True
        return self

    def order_by(self, *props, **kwargs):
        """
        Order by properties. Prepend with minus to do descending. Pass None to
        remove ordering.
        """
        pass_validation = kwargs.get('pass_validation', False)
        should_remove = len(props) == 1 and props[0] is None
        if not hasattr(self, '_order_by') or should_remove:
            self._order_by = []
            if should_remove:
                return self
        if '?' in props:
            self._order_by.append('?')
        else:
            for prop in props:
                prop = prop.strip()
                if prop.startswith('-'):
                    prop = prop[1:]
                    desc = True
                else:
                    desc = False

                if not pass_validation and prop not in self.source_class.defined_properties(
                        rels=False):
                    raise ValueError("No such property {} on {}".format(
                        prop, self.source_class.__name__))

                if not pass_validation:
                    property_obj = getattr(self.source_class, prop)
                    if isinstance(property_obj, AliasProperty):
                        prop = property_obj.aliased_to()
                self._order_by.append({
                    "name": prop,
                    "type": ' DESC ' if desc else '',
                    "ident": not pass_validation,
                })
        return self


@update_matches.register(NodeSet)
@update_matches.register(StructuredNode)
def _um_register(argument, builder, key, definition, *args, **kwargs):
    not_condition = {}
    if isinstance(argument, NodeSet) and args:
        not_condition = {'is_not': args[0]}
    builder.extra_match[key] = {
        'definition': definition,
        'type': type(argument),
        'value': argument,
        'operation': ' OR ' if 'operation' in kwargs and kwargs['operation'] == '|' else ' AND ',
        **not_condition,
    }


@generate_label.register(tuple)
@generate_label.register(list)
def _generate_label_iterable_collections(type, **kwargs):
    self = kwargs['self']
    # if in `has()` argument type is list or tuple
    labels = (kwargs['label'] + ' { uid: "' + value.uid + '"}' for value in kwargs['value'])
    where_relation = [_rel_helper(
        lhs=kwargs['source_ident'],
        rhs=label,
        ident='',
        **kwargs['val']) for label in labels]

    # add where
    self._ast['where'] += [where_relation, ]
    return self


@generate_label.register(NodeSet)
def _generate_label_ns(type, **kwargs):
    # if in `has()` argument type is `NodeSet`
    # query_params, build_query()
    is_not_linked = kwargs['is_not']
    self = kwargs['self']

    # NodeSet query builder (in `has` function)
    inner_query_builder = type.query_builder
    inner_ident = inner_query_builder.node_set.ident
    inner_query = inner_query_builder.build_query('WITH')

    # insert into origin nodeset `match` option
    matches = inner_query_builder._ast.setdefault('match', [])

    match_relation = _rel_helper(
        rhs=kwargs['source_ident'],
        lhs=inner_ident,
        reverse=True,
        **kwargs['val'])

    # add lookup
    self._ast.setdefault('lookup', [])
    self._ast['lookup'].append(inner_query)

    # add connection
    if not is_not_linked:
        self._ast['match'].append(match_relation)
    else:
        self._ast['where'].append("NOT EXISTS (" + match_relation + ")")

    # union query params of two nodesets
    self._query_params.update(inner_query_builder._query_params)
    return self


class Traversal(BaseSet):
    """
    Models a traversal from a node to another.

    :param source: Starting of the traversal.
    :type source: A :class:`~neomodel.core.StructuredNode` subclass, an
                  instance of such, a :class:`~neomodel.match.NodeSet` instance
                  or a :class:`~neomodel.match.Traversal` instance.
    :param name: A name for the traversal.
    :type name: :class:`str`
    :param definition: A relationship definition that most certainly deserves
                       a documentation here.
    :type defintion: :class:`dict`
    """

    def __init__(self, source, name, definition):
        """
        Create a traversal

        """
        super(Traversal, self).__init__()
        self.source = source

        if isinstance(source, Traversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, StructuredNode):
            self.source_class = source
        elif isinstance(source, StructuredNode):
            self.source_class = source.__class__
        elif isinstance(source, NodeSet):
            self.source_class = source.source_class
        else:
            raise TypeError("Bad source for traversal: "
                            "{}".format(type(source)))

        invalid_keys = (
            set(definition) - {'direction', 'model', 'node_class', 'relation_type'}
        )
        if invalid_keys:
            raise ValueError(
                'Unallowed keys in Traversal definition: {invalid_keys}'
                .format(invalid_keys=invalid_keys)
            )

        self.definition = definition
        self.target_class = definition['node_class']
        self.name = name
        self.rel_ident = name
        self.filters = []
        self.no_label_in_ident = False

    def match(self, **kwargs):
        """
        Traverse relationships with properties matching the given parameters.

            e.g: `.match(price__lt=10)`

        :param kwargs: see `NodeSet.filter()` for syntax
        :return: self
        """
        if kwargs:
            if self.definition.get('model') is None:
                raise ValueError("match() with filter only available on relationships with a model")
            output = process_filter_args(self.definition['model'], kwargs)
            if output:
                self.filters.append(output)
        return self


class TraversalRelationships(Traversal):

    REL_IDENT = "static_ident_rel"

    def __init__(self, source, name, definition):
        super().__init__(source, name, definition)
        self.rel_ident = self.REL_IDENT
        self.name = name
        self.target_class = definition['model']



import logging
import os
import sys
import time
import warnings

from threading import local
from functools import singledispatch
from datetime import datetime, date
from dateutil.tz import tzlocal

from neo4j.v1 import GraphDatabase, basic_auth, CypherError, SessionError, Node
from neomodel.properties import DateProperty, DateTimeProperty

from . import config
from .exceptions import UniqueProperty, ConstraintValidationFailed, ModelDefinitionMismatch

if sys.version_info >= (3, 0):
    from urllib.parse import urlparse
else:
    from urlparse import urlparse  # noqa

logger = logging.getLogger(__name__)


# make sure the connection url has been set prior to executing the wrapped function
def ensure_connection(func):
    def wrapper(self, *args, **kwargs):
        # Sort out where to find url
        if hasattr(self, 'db'):
            _db = self.db
        else:
            _db = self

        if not _db.url:
            _db.set_connection(config.DATABASE_URL)
        return func(self, *args, **kwargs)

    return wrapper


def change_neo4j_password(db, new_password):
    db.cypher_query("CALL dbms.changePassword({password})", {'password': new_password})


def clear_neo4j_database(db):
    db.cypher_query("MATCH (a) DETACH DELETE a")


class Database():
    """
    A singleton object via which all operations from neomodel to the Neo4j backend are handled with.
    """

    def __init__(self):
        """
        """
        self._active_transaction = None
        self.url = None
        self.driver = None
        self._pid = None
        # Maintains a lookup directory that is used by cypher_query
        # to infer which class to instantiate by examining the labels of the
        # node in the resultset.
        # _NODE_CLASS_REGISTRY is populated automatically by the constructor
        # of the NodeMeta type.
        self._NODE_CLASS_REGISTRY = {}

    def set_connection(self, url):
        """
        Sets the connection URL to the address a Neo4j server is set up at
        """
        u = urlparse(url)

        if u.netloc.find('@') > -1 and (u.scheme == 'bolt' or u.scheme == 'bolt+routing'):
            credentials, hostname = u.netloc.rsplit('@', 1)
            username, password, = credentials.split(':')
        else:
            raise ValueError("Expecting url format: bolt://user:password@localhost:7687"
                             " got {}".format(url))

        self.driver = GraphDatabase.driver(u.scheme + '://' + hostname,
                                           auth=basic_auth(username, password),
                                           encrypted=config.ENCRYPTED_CONNECTION,
                                           max_pool_size=config.MAX_POOL_SIZE)
        self.url = url
        self._pid = os.getpid()
        self._active_transaction = None

    @property
    def transaction(self):
        """
        Returns the current transaction object
        """
        return TransactionProxy(self)

    @property
    def write_transaction(self):
        return TransactionProxy(self, access_mode="WRITE")

    @property
    def read_transaction(self):
        return TransactionProxy(self, access_mode="READ")

    @ensure_connection
    def begin(self, access_mode=None):
        """
        Begins a new transaction, raises SystemError exception if a transaction is in progress
        """
        if self._active_transaction:
            raise SystemError("Transaction in progress")
        self._active_transaction = self.driver.session(access_mode=access_mode).begin_transaction()

    @ensure_connection
    def commit(self):
        """
        Commits the current transaction
        """
        r = self._active_transaction.commit()
        self._active_transaction = None
        return r

    @ensure_connection
    def rollback(self):
        """
        Rolls back the current transaction
        """
        self._active_transaction.rollback()
        self._active_transaction = None

    def _object_resolution(self, result_list):
        """
        Performs in place automatic object resolution on a set of results
        returned by cypher_query.

        The function operates recursively in order to be able to resolve Nodes
        within nested list structures. Not meant to be called directly,
        used primarily by cypher_query.

        :param result_list: A list of results as returned by cypher_query.
        :type list:

        :return: A list of instantiated objects.
        """

        # Object resolution occurs in-place
        for a_result_item in enumerate(result_list):
            for a_result_attribute in enumerate(a_result_item[1]):
                try:
                    # Primitive types should remain primitive types,
                    #  Nodes to be resolved to native objects
                    resolved_object = a_result_attribute[1]

                    if isinstance(a_result_attribute[1], Node):
                        resolved_object = self._NODE_CLASS_REGISTRY[frozenset(
                            a_result_attribute[1].labels)].inflate(a_result_attribute[1])

                    if isinstance(a_result_attribute[1], list):
                        resolved_object = self._object_resolution([a_result_attribute[1]])

                    result_list[a_result_item[0]][a_result_attribute[0]] = resolved_object

                except KeyError:
                    # Not being able to match the label set of a node with a known object results
                    # in a KeyError in the internal dictionary used for resolution. If it is impossible
                    # to match, then raise an exception with more details about the error.
                    raise ModelDefinitionMismatch(a_result_attribute[1], self._NODE_CLASS_REGISTRY)

        return result_list

    @ensure_connection
    def cypher_query(
            self,
            query,
            params=None,
            handle_unique=True,
            retry_on_session_expire=False,
            resolve_objects=False):
        """
        Runs a query on the database and returns a list of results and their headers.

        :param query: A CYPHER query
        :type: str
        :param params: Dictionary of parameters
        :type: dict
        :param handle_unique: Whether or not to raise UniqueProperty exception on Cypher's ConstraintValidation errors
        :type: bool
        :param retry_on_session_expire: Whether or not to attempt the same query again if the transaction has expired
        :type: bool
        :param resolve_objects: Whether to attempt to resolve the returned nodes to data model objects automatically
        :type: bool
        """

        if self._pid != os.getpid():
            self.set_connection(self.url)

        if self._active_transaction:
            session = self._active_transaction
        else:
            session = self.driver.session()

        try:
            # Retrieve the data
            start = time.clock()
            response = session.run(query, params)
            results, meta = [list(r.values()) for r in response], response.keys()
            end = time.clock()

            if resolve_objects:
                # Do any automatic resolution required
                results = self._object_resolution(results)

        except CypherError as ce:
            if ce.code == u'Neo.ClientError.Schema.ConstraintValidationFailed':
                if 'already exists with label' in ce.message and handle_unique:
                    raise UniqueProperty(ce.message)

                raise ConstraintValidationFailed(ce.message)
            else:
                exc_info = sys.exc_info()
                if sys.version_info >= (3, 0):
                    raise exc_info[1].with_traceback(exc_info[2])
                else:
                    raise exc_info[1]
        except SessionError:
            if retry_on_session_expire:
                self.set_connection(self.url)
                return self.cypher_query(query=query,
                                         params=params,
                                         handle_unique=handle_unique,
                                         retry_on_session_expire=False)
            raise
        if os.environ.get('NEOMODEL_CYPHER_DEBUG', '0') == '1':
            logger.debug(
                "query: " +
                query +
                "\nparams: " +
                repr(params) +
                "\ntook: %.2gs\n" %
                (end -
                 start))

        return results, meta


class TransactionProxy(object):
    def __init__(self, db, access_mode=None):
        self.db = db
        self.access_mode = access_mode

    @ensure_connection
    def __enter__(self):
        self.db.begin(access_mode=self.access_mode)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value:
            self.db.rollback()

        if exc_type is CypherError:
            if exc_value.code == u'Neo.ClientError.Schema.ConstraintValidationFailed':
                raise UniqueProperty(exc_value.message)

        if not exc_value:
            self.db.commit()

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper


def deprecated(message):
    def f__(f):
        def f_(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)

        f_.__name__ = f.__name__
        f_.__doc__ = f.__doc__
        f_.__dict__.update(f.__dict__)
        return f_

    return f__


def classproperty(f):
    class cpf(object):
        def __init__(self, getter):
            self.getter = getter

        def __get__(self, obj, type=None):
            return self.getter(type)

    return cpf(f)


# Just used for error messages
class _UnsavedNode(object):
    def __repr__(self):
        return '<unsaved node>'

    def __str__(self):
        return self.__repr__()


class UnionBlock(object):

    def __init__(self, source, field, field_definition, nodeset, **kwargs):
        """
        Attrs:
            source (<NodeSet>): source  model
            field (str): the relationship field name from source model
            nodeset (<BaseSet>): nodeset
        """
        self.source = source
        self.field = field
        self.field_definition = field_definition
        self.nodeset = nodeset
        self.source_model = kwargs.get('source_model')


    def build_query(self, ):
        """
        Returns:
            query: str
            params: dict
        """
        relationship = {
            self.field: self.nodeset
        }
        qb = self.source_model.nodes.has(**relationship).query_builder
        query = qb.build_query()
        params = qb._query_params
        return query, params

    @staticmethod
    def union_queries(source_ident, union_blocks):
        query_params = map(lambda x: x.build_query(), union_blocks)
        queries = []
        parameters = {}
        for query, params in query_params:
            queries.append(query)
            parameters.update(params)
        query = " UNION ".join(queries)
        parameter = "union_block"
        params = {parameter: parameters}

        # only python 3.6 support
        pattern = f"""
        CALL apoc.cypher.run(
            "{query}",
            ${parameter}
        ) YIELD value
        WITH value.{source_ident} as {source_ident}
        """
        return pattern, params


def _get_node_properties(node):
    """Get the properties from a neo4j.v1.types.graph.Node object."""
    # 1.6.x and newer have it as `_properties`
    if hasattr(node, '_properties'):
        return node._properties
    # 1.5.x and older have it as `properties`
    else:
        return node.properties


@singledispatch
def auto_update_field(field, node, key):
    raise NotImplementedError("{} not implemented yet.".format(type(field)))


@auto_update_field.register(DateTimeProperty)
def auf_datetimeproperty(field, node, key):
    setattr(node, key, datetime.now(tzlocal()))
    return node


@auto_update_field.register(DateProperty)
def auf_datetimeproperty(field, node, key):
    setattr(node, key, datetime.now(tzlocal()).date())
    return node


def auto_update(node):
    """ Auto update hook (for update magic dates in models)
    """
    for k, v in node.defined_properties(aliases=False, rels=False).items():
        if isinstance(v, (DateProperty, DateTimeProperty)):
            if v.auto_update:
                auto_update_field(v, node, k)

def get_rhs_ident(rel_field):
    return rel_field._raw_class.split('.')[-1]



from neomodel import StructuredNode, StringProperty, db
from multiprocessing import Pool


class ThingyMaBob(StructuredNode):
    name = StringProperty(unique_index=True, required=True)


def thing_create(name):
    name = str(name)
    thing, = ThingyMaBob.get_or_create({'name': name})
    return thing.name, name


def test_concurrency():
    p = Pool(5)
    results = p.map(thing_create, range(50))
    for returned, sent in results:
        assert returned == sent



from __future__ import print_function
import warnings
import os
import sys

from neomodel import config, db, clear_neo4j_database, change_neo4j_password
from neo4j.v1 import CypherError


def pytest_addoption(parser):
    """
    Adds the command line option --resetdb.

    :param parser: The parser object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption>`_
    :type Parser object: For more information please see <https://docs.pytest.org/en/latest/reference.html#_pytest.config.Parser>`_
    """
    parser.addoption("--resetdb", action="store_true", help = "Ensures that the database is clear prior to running tests for neomodel", default=False)

def pytest_sessionstart(session):
    """
    Provides initial connection to the database and sets up the rest of the test suite

    :param session: The session object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_sessionstart>`_
    :type Session object: For more information please see <https://docs.pytest.org/en/latest/reference.html#session>`_
    """

    warnings.simplefilter('default')

    config.DATABASE_URL = os.environ.get('NEO4J_BOLT_URL', 'bolt://neo4j:neo4j_admin@localhost:7687')
    config.AUTO_INSTALL_LABELS = True

    try:
        # Clear the database if required
        database_is_populated, _ = db.cypher_query("MATCH (a) return count(a)>0 as database_is_populated")
        if database_is_populated[0][0] and not session.config.getoption("resetdb"):
            raise SystemError("Please note: The database seems to be populated.\n\tEither delete all nodes and edges manually, or set the --resetdb parameter when calling pytest\n\n\tpytest --resetdb.")
        else:
            clear_neo4j_database(db)
    except CypherError as ce:
        # Handle instance without password being changed
        if 'The credentials you provided were valid, but must be changed before you can use this instance' in str(ce):
            warnings.warn("New database with no password set, setting password to 'test'")
            change_neo4j_password(db, 'test')
            db.set_connection('bolt://neo4j:neo4j_admin@localhost:7687')
            warnings.warn("Please 'export NEO4J_BOLT_URL=bolt://neo4j:neo4j_admin@localhost:7687' for subsequent test runs")
        else:
            raise ce



import sys

from setuptools import setup, find_packages

setup(
    name='neomodel-next',
    version='3.4.0',
    description='An object mapper for the neo4j graph database.',
    long_description=open('README.rst').read(),
    author='Robin Edwards; Mardanov Timur',
    author_email='timurmardanov97@gmail.com',
    zip_safe=True,
    url='http://github.com/MardanovTimur/neomodel-next',
    license='MIT',
    packages=find_packages(exclude=('tests',)),
    keywords='graph neo4j neomodel ORM OGM',
    scripts=['scripts/neomodel_install_labels', 'scripts/neomodel_remove_labels'],
    setup_requires=['pytest-runner'] if any(x in ('pytest', 'test') for x in sys.argv) else [],
    tests_require=['pytest'],
    install_requires=[
        'neo4j-driver>=1.5.2, <1.7.0',
        'pytz>=2016.10',
        'python-dateutil==2.8.0',
        'numpy>=1.12.1',
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Programming Language :: Python :: 3.6",
        "Topic :: Database",
    ])



import functools
import json
import sys
import types
import re
import uuid
import warnings
import numpy as np
import pytz

from datetime import date, datetime
# for local timezone
from dateutil.tz import tzlocal

from neomodel import config
from neomodel.exceptions import InflateError, DeflateError, RequiredProperty


if sys.version_info >= (3, 0):
    unicode = str


def display_for(key):
    def display_choice(self):
        return getattr(self.__class__, key).choices[getattr(self, key)]
    return display_choice


class PropertyManager(object):
    """
    Common methods for handling properties on node and relationship objects.
    """

    def __init__(self, **kwargs):
        properties = getattr(self, "__all_properties__", None)
        if properties is None:
            properties = \
                self.defined_properties(rels=False, aliases=False).items()
        for name, property in properties:
            if kwargs.get(name) is None:
                if getattr(property, 'has_default', False):
                    setattr(self, name, property.default_value())
                else:
                    setattr(self, name, None)
            else:
                setattr(self, name, kwargs[name])

            if getattr(property, 'choices', None):
                setattr(self, 'get_{}_display'.format(name),
                        types.MethodType(display_for(name), self))

            if name in kwargs:
                del kwargs[name]

        aliases = getattr(self, "__all_aliases__", None)
        if aliases is None:
            aliases = self.defined_properties(
                aliases=True, rels=False, properties=False).items()
        for name, property in aliases:
            if name in kwargs:
                setattr(self, name, kwargs[name])
                del kwargs[name]

        # undefined properties (for magic @prop.setters etc)
        for name, property in kwargs.items():
            setattr(self, name, property)

    @property
    def __properties__(self):
        from .relationship_manager import RelationshipManager

        return dict((name, value) for name, value in vars(self).items()
                    if not name.startswith('_')
                    and not callable(value)
                    and not isinstance(value,
                                       (RelationshipManager, AliasProperty,))
                    )

    @classmethod
    def deflate(cls, properties, obj=None, skip_empty=False):
        # deflate dict ready to be stored
        deflated = {}
        for name, property \
                in cls.defined_properties(aliases=False, rels=False).items():
            db_property = property.db_property or name
            if properties.get(name) is not None:
                deflated[db_property] = property.deflate(properties[name], obj)
            elif property.has_default:
                deflated[db_property] = property.deflate(
                    property.default_value(), obj
                )
            elif property.required or property.unique_index:
                raise RequiredProperty(name, cls)
            elif not skip_empty:
                deflated[db_property] = None
        return deflated

    @classmethod
    def defined_properties(cls, aliases=True, properties=True, rels=True):
        from .relationship_manager import RelationshipDefinition, RelationshipManager
        props = {}
        for baseclass in reversed(cls.__mro__):
            props.update(dict(
                (name, property) for name, property in vars(baseclass).items()
                if (aliases and isinstance(property, AliasProperty))
                or (properties and isinstance(property, Property)
                    and not isinstance(property, AliasProperty))
                or (rels and isinstance(property, (RelationshipDefinition, RelationshipManager)))
            ))
        return props


def validator(fn):
    fn_name = fn.func_name if hasattr(fn, 'func_name') else fn.__name__
    if fn_name == 'inflate':
        exc_class = InflateError
    elif fn_name == 'deflate':
        exc_class = DeflateError
    else:
        raise Exception("Unknown Property method " + fn_name)

    @functools.wraps(fn)
    def _validator(self, value, obj=None, rethrow=True):
        if rethrow:
            try:
                return fn(self, value)
            except Exception as e:
                raise exc_class(self.name, self.owner, str(e), obj)
        else:
            # For using with ArrayProperty where we don't want an Inflate/Deflate error.
            return fn(self, value)

    return _validator


class Property(object):
    """
    Base class for object properties.

    :param unique_index: Creates a unique index for this property. Defaults to
                         ``False``.
    :type unique_index: :class:`bool`
    :param index: Creates an index for this property. Defaults to ``False``.
    :type index: :class:`bool`
    :param required: Marks the property as required. Defaults to ``False``.
    :type required: :class:`bool`
    :param default: A default value or callable that returns one to set when a
                    node is initialized without specifying this property.
    :param db_property: A name that this property maps to in the database.
                        Defaults to the model's property name.
    :type db_property: :class:`str`
    :param label: Optional, used by ``django_neomodel``.
    :type label: :class:`str`
    :param help_text: Optional, used by ``django_neomodel``.
    :type help_text: :class:`str`
    """

    form_field_class = 'CharField'

    def __init__(self, unique_index=False, index=False, required=False, default=None,
                 db_property=None, label=None, help_text=None, **kwargs):

        if default is not None and required:
            raise ValueError(
                "The arguments `required` and `default` are mutually exclusive."
            )
        if unique_index and index:
            raise ValueError(
                "The arguments `unique_index` and `index` are mutually exclusive."
            )

        self.required = required
        self.unique_index = unique_index
        self.index = index
        self.default = default
        self.has_default = True if self.default is not None else False
        self.db_property = db_property
        self.label = label
        self.help_text = help_text

    def default_value(self):
        """
        Generate a default value

        :return: the value
        """
        if self.has_default:
            if hasattr(self.default, '__call__'):
                return self.default()
            else:
                return self.default
        else:
            raise Exception("No default value specified")

    @property
    def is_indexed(self):
        return self.unique_index or self.index


class NormalizedProperty(Property):
    """
    Base class for normalized properties. These use the same normalization
    method to in- or deflating.
    """

    @validator
    def inflate(self, value):
        return self.normalize(value)

    @validator
    def deflate(self, value):
        return self.normalize(value)

    def default_value(self):
        default = super(NormalizedProperty, self).default_value()
        return self.normalize(default)

    def normalize(self, value):
        raise NotImplementedError('Specialize normalize method')


# TODO remove this with the next major release
def _warn_NormalProperty_renamed():
    warnings.warn(
        'The class NormalProperty was renamed to NormalizedProperty. '
        'Use that one as base class. The former will be removed in the next '
        'major release.', DeprecationWarning)


if sys.version_info >= (3, 6):
    class NormalProperty(NormalizedProperty):
        def __init_subclass__(cls, **kwargs):
            _warn_NormalProperty_renamed()
else:
    class NormalProperty(NormalizedProperty):
        def __init__(self, *args, **kwargs):
            _warn_NormalProperty_renamed()
            super(NormalProperty, self).__init__(*args, **kwargs)
##


class RegexProperty(NormalizedProperty):
    """
    Validates a property against a regular expression.

    If sub-classing set:

        expression = r'[^@]+@[^@]+\.[^@]+'
    """
    form_field_class = 'RegexField'

    expression = None

    def __init__(self, expression=None, **kwargs):
        """
        Initializes new property with an expression.

        :param str expression: regular expression validating this property
        """
        super(RegexProperty, self).__init__(**kwargs)
        actual_re = expression or self.expression
        if actual_re is None:
            raise ValueError('expression is undefined')
        self.expression = actual_re

    def normalize(self, value):
        normal = unicode(value)
        if not re.match(self.expression, normal):
            raise ValueError(
                '{0!r} does not matches {1!r}'.format(
                    value,
                    self.expression,
                )
            )
        return normal


class EmailProperty(RegexProperty):
    """
    Store email addresses
    """
    form_field_class = 'EmailField'
    expression = r'[^@]+@[^@]+\.[^@]+'


class StringProperty(NormalizedProperty):
    """
    Stores a unicode string

    :param choices: A mapping of valid strings to label strings that are used
                    to display information in an application. If the default
                    value ``None`` is used, any string is valid.
    :type choices: Any type that can be used to initiate a :class:`dict`.
    """

    def __init__(self, choices=None, **kwargs):
        super(StringProperty, self).__init__(**kwargs)

        if choices is None:
            self.choices = None
        else:
            try:
                self.choices = dict(choices)
            except Exception:
                raise ValueError("The choices argument must be convertable to "
                                 "a dictionary.")
            # Python 3:
            # except Exception as e:
            #     raise ValueError("The choices argument must be convertable to "
            #                      "a dictionary.") from e
            self.form_field_class = 'TypedChoiceField'

    def normalize(self, value):
        if self.choices is not None and value not in self.choices:
            raise ValueError("Invalid choice: {}".format(value))
        return unicode(value)

    def default_value(self):
        return self.normalize(super(StringProperty, self).default_value())


class IntegerProperty(Property):
    """
    Stores an Integer value
    """
    form_field_class = 'IntegerField'

    @validator
    def inflate(self, value):
        return int(value)

    @validator
    def deflate(self, value):
        return int(value)

    def default_value(self):
        return int(super(IntegerProperty, self).default_value())


class ArrayProperty(Property):
    """
    Stores a list of items
    """

    def __init__(self, base_property=None, **kwargs):
        """
        Store a list of values, optionally of a specific type.

        :param base_property: List item type e.g StringProperty for string
        :type: Property
        """

        # list item type
        if base_property is not None:
            if not isinstance(base_property, Property):
                raise TypeError('Expecting neomodel Property')

            if isinstance(base_property, ArrayProperty):
                raise TypeError('Cannot have nested ArrayProperty')

            for ilegal_attr in ['default', 'index', 'unique_index', 'required']:
                if getattr(base_property, ilegal_attr, None):
                    raise ValueError(
                        'ArrayProperty base_property cannot have "{}" set'.format(ilegal_attr))

        self.base_property = base_property

        super(ArrayProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        if self.base_property:
            return [self.base_property.inflate(item, rethrow=False) for item in value]

        return list(value)

    @validator
    def deflate(self, value):
        if self.base_property:
            return [self.base_property.deflate(item, rethrow=False) for item in value]

        return list(value)

    def default_value(self):
        return list(super(ArrayProperty, self).default_value())


class FloatProperty(Property):
    """
    Store a floating point value
    """
    form_field_class = 'FloatField'

    @validator
    def inflate(self, value):
        return float(value)

    @validator
    def deflate(self, value):
        return float(value)

    def default_value(self):
        return float(super(FloatProperty, self).default_value())


class BooleanProperty(Property):
    """
    Stores a boolean value
    """
    form_field_class = 'BooleanField'

    @validator
    def inflate(self, value):
        return bool(value)

    @validator
    def deflate(self, value):
        return bool(value)

    def default_value(self):
        return bool(super(BooleanProperty, self).default_value())


class DateProperty(Property):
    """
    Stores a date
    """
    form_field_class = 'DateField'

    def __init__(self, auto_update=False, **kwargs):
        """
        Args:
            auto_update (bool), default = `False`: auto update time,
                when object is updated
        """
        self.auto_update = auto_update
        super(DateProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        return datetime.strptime(unicode(value), "%Y-%m-%d").date()

    @validator
    def deflate(self, value):
        if not isinstance(value, date):
            msg = 'datetime.date object expected, got {0}'.format(repr(value))
            raise ValueError(msg)
        return value.isoformat()


class DateTimeProperty(DateProperty, Property):
    """ A property representing a :class:`datetime.datetime` object as
        unix epoch.

        :param default_now: If ``True``, the creation time (UTC) will be used as default.
                            Defaults to ``False``.
        :type default_now: :class:`bool`
    """
    form_field_class = 'DateTimeField'

    def __init__(self, default_now=False, **kwargs):
        if default_now:
            if 'default' in kwargs:
                raise ValueError('too many defaults')
            kwargs['default'] = lambda: datetime.now(tzlocal())
        super(DateTimeProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        try:
            epoch = float(value)
        except ValueError:
            raise ValueError("Float or integer expected, got {0} can't inflate to "
                             "datetime.".format(type(value)))
        return datetime.fromtimestamp(epoch, tzlocal())

    @validator
    def deflate(self, value):
        if not isinstance(value, datetime):
            raise ValueError('datetime object expected, got {0}.'.format(type(value)))
        if value.tzinfo:
            value = value.astimezone(pytz.utc)
            epoch_date = datetime(1970, 1, 1, tzinfo=pytz.utc)
        elif config.FORCE_TIMEZONE:
            raise ValueError("Error deflating {}: No timezone provided.".format(value))
        else:
            # No timezone specified on datetime object.. assuming UTC
            epoch_date = datetime(1970, 1, 1)
        return float((value - epoch_date).total_seconds())


class JSONProperty(Property):
    """
    Store a data structure as a JSON string.

    The structure will be inflated when a node is retrieved.
    """

    def __init__(self, *args, **kwargs):
        super(JSONProperty, self).__init__(*args, **kwargs)

    @validator
    def inflate(self, value):
        return json.loads(value)

    @validator
    def deflate(self, value):
        return json.dumps(value)


class JsonArrayProperty(Property):
    """Stored in neo4j as json format
    Checks : np.ndarray, list, tuple

    Arguments:
        shape (int), default=1: shape of array
    """

    def __init__(self, *args, **kwargs):
        self.type = kwargs.pop('type', str)
        super(JsonArrayProperty, self).__init__(*args, **kwargs)

    @validator
    def inflate(self, value):
        return json.loads(value)

    def iterate_inn(self, arr):
        for i in arr:
            if isinstance(i, (list, np.ndarray, tuple)):
                self.iterate_inn(i)
            else:
                if not isinstance(i, self.type):
                    raise Exception("Incorrect type {i}".format(i=i))

    @validator
    def deflate(self, value):
        if not isinstance(value, (list, tuple, np.ndarray)):
            raise Exception("This type should be list or np.ndarray")
        arr = np.array(value)
        self.iterate_inn(arr)
        return json.dumps(value)


class AliasProperty(property, Property):
    """
    Alias another existing property
    """

    def __init__(self, to=None):
        """
        Create new alias

        :param to: name of property aliasing
        :type: str
        """
        self.target = to
        self.required = False
        self.has_default = False

    def aliased_to(self):
        return self.target

    def __get__(self, obj, cls):
        return getattr(obj, self.aliased_to()) if obj else self

    def __set__(self, obj, value):
        setattr(obj, self.aliased_to(), value)

    @property
    def index(self):
        return getattr(self.owner, self.aliased_to()).index

    @property
    def unique_index(self):
        return getattr(self.owner, self.aliased_to()).unique_index


class UniqueIdProperty(Property):
    """
    A unique identifier, a randomly generated uid (uuid4) with a unique index
    """

    def __init__(self, **kwargs):
        for item in ['required', 'unique_index', 'index', 'default']:
            if item in kwargs:
                raise ValueError('{} argument ignored by {}'.format(item, self.__class__.__name__))

        kwargs['unique_index'] = True
        kwargs['default'] = lambda: uuid.uuid4().hex
        super(UniqueIdProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value):
        return unicode(value)

    @validator
    def deflate(self, value):
        return unicode(value)



from neo4j.addressing import AddressError
from pytest import raises

from neomodel import db, StructuredNode, StringProperty, UniqueProperty


class Person(StructuredNode):
    name = StringProperty(unique_index=True)


def test_rollback_and_commit_transaction():
    for p in Person.nodes:
        p.delete()

    Person(name='Roger').save()

    db.begin()
    Person(name='Terry S').save()
    db.rollback()

    assert len(Person.nodes) == 1

    db.begin()
    Person(name='Terry S').save()
    db.commit()

    assert len(Person.nodes) == 2


@db.transaction
def in_a_tx(*names):
    for n in names:
        Person(name=n).save()


def test_transaction_decorator():
    for p in Person.nodes:
        p.delete()

    # should work
    in_a_tx('Roger')
    assert True

    # should bail but raise correct error
    with raises(UniqueProperty):
        in_a_tx('Jim', 'Roger')

    assert 'Jim' not in [p.name for p in Person.nodes]


def test_transaction_as_a_context():
    with db.transaction:
        Person(name='Tim').save()

    assert Person.nodes.filter(name='Tim')

    with raises(UniqueProperty):
        with db.transaction:
            Person(name='Tim').save()


def test_query_inside_transaction():
    for p in Person.nodes:
        p.delete()

    with db.transaction:
        Person(name='Alice').save()
        Person(name='Bob').save()

        assert len([p.name for p in Person.nodes]) == 2


def test_set_connection_works():
    assert Person(name='New guy').save()
    from socket import gaierror

    old_url = db.url
    with raises(AddressError):
        db.set_connection('bolt://user:password@nowhere:7687')
    db.set_connection(old_url)
    # set connection back
    assert Person(name='New guy2').save()
